@@ -7,16 +7,19 @@
     unreachable_pub
 )]
 
+use parking_lot::RwLock;
 use std::borrow::Cow;
 use std::collections::{HashMap, HashSet};
 use std::fmt::{Debug, Formatter};
-use std::io::{stderr, Write};
+use std::io::{ stderr, Write};
 use std::ops::DerefMut;
-use std::sync::{Arc, Mutex, RwLock};
+use std::sync::{Arc, Mutex};
 use std::time::{Duration, SystemTime};
+use parking_lot::lock_api::RwLockUpgradableReadGuard;
 use tracing::field::{Field, Visit};
 use tracing::span::{Attributes, Record};
 
+use tracing::subscriber::set_global_default;
 use tracing::{span, Event as TracingEvent, Id, Span, Subscriber};
 use tracing_subscriber::field::RecordFields;
 
@@ -25,6 +28,7 @@ use tracing_subscriber::registry::LookupSpan;
 use tracing_subscriber::{Layer, Registry};
 
 const NESTED_EVENT_OFFSET: usize = 2;
+const DURATION_WIDTH: usize = 6;
 
 #[derive(Debug, Clone)]
 struct EventInfo {
@@ -35,7 +39,7 @@ struct EventInfo {
 impl EventInfo {
     fn to_string(&self, settings: &Settings) -> String {
         let mut out = String::new();
-        self.metadata.write(&mut out, settings).unwrap();
+        self.metadata.write(&mut out, settings).expect("writing to a string cannot fail");
         out
     }
 }
@@ -46,7 +50,10 @@ impl SpanInfo {
         S: Subscriber + for<'span> LookupSpan<'span> + Send + Sync,
     {
         Self {
-            name: ctx.metadata(span).unwrap().name().to_string(),
+            name: ctx
+                .metadata(span)
+                .map(|metadata| metadata.name())
+                .unwrap_or("could-not-find-span"),
             start: SystemTime::now(),
             end: None,
         }
@@ -125,11 +132,13 @@ impl TrackedMetadata {
         }
 
         write!(f, "{{")?;
-        for (k, v) in relevant_fields().take(self.data.len() - 1) {
-            write!(f, "{}: {}, ", k, v)?;
+        let mut peekable = relevant_fields().peekable();
+        while let Some((k, v)) = peekable.next() {
+            write!(f, "{}: {}", k, v)?;
+            if peekable.peek().is_some() {
+                write!(f, ", ")?;
+            }
         }
-        let (k, v) = relevant_fields().last().expect("non empty");
-        write!(f, "{}: {}", k, v)?;
         write!(f, "}}")
     }
 }
@@ -138,7 +147,7 @@ impl TrackedMetadata {
 struct SpanInfo {
     start: SystemTime,
     end: Option<SystemTime>,
-    name: String,
+    name: &'static str,
 }
 
 impl SpanInfo {
@@ -151,6 +160,10 @@ impl SpanInfo {
         id
     }
 
+    fn duration(&self) -> Option<Duration> {
+        self.end.and_then(|end|end.duration_since(self.start).ok())
+    }
+
     fn render(
         &self,
         out: &mut dyn Write,
@@ -163,11 +176,10 @@ impl SpanInfo {
         let truncated_key_width = render_conf.key_width - left_offset;
         key.truncate(truncated_key_width);
         let ev_start_ts = self.start;
-        let ev_end_ts = match self.end {
+        let span_len = match self.duration() {
             None => return Ok(()),
-            Some(ts) => ts,
+            Some(dur) => dur,
         };
-        let span_len = ev_end_ts.duration_since(ev_start_ts).unwrap();
         if let Some(min_duration) = settings.min_duration.as_ref() {
             if &span_len < min_duration {
                 return Ok(());
@@ -181,15 +193,15 @@ impl SpanInfo {
             out,
             " {:>dur_width$} ",
             pretty_duration(span_len),
-            dur_width = 6
+            dur_width = DURATION_WIDTH
         )?;
 
         let offset = width(
             render_conf.chart_width(),
             render_conf.total(),
-            ev_start_ts.duration_since(render_conf.start_ts).unwrap(),
+            match ev_start_ts.duration_since(render_conf.start_ts) { Ok(dur) => dur, Err(_) => return Ok(())},
         );
-        write!(out, "{}", " ".repeat(offset)).unwrap();
+        write!(out, "{}", " ".repeat(offset))?;
         let interval_width = width(render_conf.chart_width(), render_conf.total(), span_len);
         match interval_width {
             0 => write!(out, "┆"),
@@ -236,13 +248,14 @@ struct RenderConf {
 
 impl RenderConf {
     fn total(&self) -> Duration {
+        // start_ts is always less than end_ts
         self.end_ts.duration_since(self.start_ts).unwrap()
     }
 
     fn chart_width(&self) -> usize {
         self.width
             .checked_sub(self.key_width)
-            .and_then(|w| w.checked_sub(8))
+            .and_then(|w| w.checked_sub(DURATION_WIDTH + 2))
             .unwrap_or(20)
     }
 }
@@ -301,21 +314,19 @@ impl SpanTracker {
 
     fn path<'a, S>(
         &self,
-        span: &span::Id,
-        ctx: &'a Context<'a, S>,
+        scope: tracing_subscriber::registry::Scope<'a, S>,
     ) -> impl Iterator<Item = span::Id> + 'a
     where
         S: Subscriber + for<'span> LookupSpan<'span> + Send + Sync,
     {
         let root = self.span_id.clone();
-        let mut path = ctx
-            .span_scope(span)
-            .unwrap()
+
+        let path = scope
             .from_root()
             .skip_while(move |span| span.id() != root)
+            // skip until the root, then skip over the root
+            .skip(1)
             .map(|s| s.id());
-        let path_root = path.next().expect("one item exists");
-        assert_eq!(&path_root, &self.span_id);
         path
     }
 
@@ -392,7 +403,7 @@ impl SpanTracker {
         }
 
         if settings.types.events {
-            let left_offset = left_offset + 1;
+            let left_offset = left_offset + 2;
             let truncated_key_width = render_conf.key_width - left_offset;
             let base_offset = width(
                 render_conf.chart_width(),
@@ -400,7 +411,7 @@ impl SpanTracker {
                 span_info
                     .start
                     .duration_since(render_conf.start_ts)
-                    .unwrap(),
+                    .expect("start_ts MUST be before span_info.start because it is a minima"),
             );
             let mut settings_with_message = settings.clone();
             if let FieldFilter::AllowList(list) = &mut settings_with_message.field_printing {
@@ -419,7 +430,7 @@ impl SpanTracker {
                     ev.timestamp.duration_since(span_info.start).unwrap(),
                 ) as i32)
                     - 1;
-                write!(out, "{}", " ".repeat(8))?;
+                write!(out, "{}", " ".repeat(DURATION_WIDTH + 2))?;
                 writeln!(
                     out,
                     "{}┼",
@@ -432,7 +443,12 @@ impl SpanTracker {
         children.sort_by_key(|child| child.info.as_ref().map(|it| it.start));
 
         for child in children {
-            child._dump(out, render_conf, settings, left_offset + 2)?;
+            child._dump(
+                out,
+                render_conf,
+                settings,
+                left_offset + NESTED_EVENT_OFFSET,
+            )?;
         }
         Ok(())
     }
@@ -470,16 +486,16 @@ pub struct Settings {
     types: Types,
     field_printing: FieldFilter,
     updated: bool,
-    out: WriterWrapper,
+    out: DynWriter,
 }
 
 /// Wrap a dyn writer to get a Debug implementation
 #[derive(Clone)]
-struct WriterWrapper {
+struct DynWriter {
     inner: Arc<Mutex<dyn std::io::Write + Send>>,
 }
 
-impl Debug for WriterWrapper {
+impl Debug for DynWriter {
     fn fmt(&self, f: &mut Formatter<'_>) -> std::fmt::Result {
         write!(f, "dyn Writer")
     }
@@ -518,7 +534,7 @@ impl Default for Settings {
             },
             field_printing: Default::default(),
             updated: false,
-            out: WriterWrapper {
+            out: DynWriter {
                 inner: Arc::new(Mutex::new(stderr())),
             },
         }
@@ -550,7 +566,7 @@ impl Settings {
 
     /// Overwrite the writer [`TexRayLayer`] will output to
     pub fn writer<W: Write + Send + 'static>(mut self, w: W) -> Self {
-        self.out = WriterWrapper {
+        self.out = DynWriter {
             inner: Arc::new(Mutex::new(w)),
         };
         self
@@ -623,6 +639,13 @@ impl TeXRayLayer {
         dumper
     }
 
+    /// Install [`TexRayLayer`] as the global default
+    pub fn init(self) {
+        let registry = tracing_subscriber::registry().with(self);
+        use tracing_subscriber::layer::SubscriberExt;
+        set_global_default(registry).expect("failed to install subscriber")
+    }
+
     fn settings(&self) -> &Settings {
         &self.settings
     }
@@ -678,15 +701,19 @@ impl TeXRayLayer {
         &self,
         span: &span::Id,
         ctx: &Context<'a, S>,
-        f: impl Fn(&mut SpanTracker),
+        f: impl Fn(&mut SpanTracker, tracing_subscriber::registry::Scope<'_, S>),
     ) where
         S: Subscriber + for<'span> LookupSpan<'span> + Send + Sync,
     {
-        if let Some(span_iter) = ctx.span_scope(span) {
-            let mut tracker = self.tracked_spans.write().unwrap();
-            for span_ref in span_iter {
-                if let Some(inner) = tracker.get_mut(&span_ref.id()) {
-                    f(inner)
+        if let Some(mut span_iter) = ctx.span_scope(span) {
+            let trackers = self.tracked_spans.upgradable_read();
+            if span_iter.any(|span| trackers.contains_key(&span.id())) {
+                let span_iter = ctx.span_scope(span).expect("span scope exists, loaded above");
+                let mut trackers = RwLockUpgradableReadGuard::upgrade(trackers);
+                for span_ref in span_iter {
+                    if let Some(span_tracker) = trackers.get_mut(&span_ref.id()) {
+                        f(span_tracker, ctx.span_scope(span).unwrap())
+                    }
                 }
             }
         }
@@ -696,7 +723,6 @@ impl TeXRayLayer {
         if let Some(id) = span.id() {
             self.tracked_spans
                 .write()
-                .unwrap()
                 .insert(id.clone(), SpanTracker::new(id, settings));
         }
     }
@@ -723,14 +749,6 @@ fn pretty_duration(duration: Duration) -> String {
     unreachable!("{:?}", duration)
 }
 
-#[cfg(test)]
-mod tests {
-    #[test]
-    fn it_works() {
-        assert_eq!(2 + 2, 4);
-    }
-}
-
 impl<S> Layer<S> for TeXRayLayer
 where
     S: Subscriber + for<'span> LookupSpan<'span> + Send + Sync,
@@ -743,35 +761,35 @@ where
                 timestamp: SystemTime::now(),
                 metadata,
             };
-            self.for_relevant_trackers(span, &ctx, |tracker| {
-                tracker.add_event(span, tracker.path(span, &ctx), tracked_event.clone())
+            self.for_relevant_trackers(span, &ctx, |tracker, scope| {
+                tracker.add_event(span, tracker.path(scope), tracked_event.clone())
             })
         }
     }
 
     fn on_new_span(&self, attrs: &Attributes<'_>, id: &Id, ctx: Context<'_, S>) {
-        self.for_relevant_trackers(id, &ctx, |tracker| {
-            tracker.record_metadata(id, tracker.path(id, &ctx), attrs)
+        self.for_relevant_trackers(id, &ctx, |tracker, scope| {
+            tracker.record_metadata(id, tracker.path(scope), attrs)
         })
     }
 
     fn on_record(&self, id: &Id, values: &Record<'_>, ctx: Context<'_, S>) {
-        self.for_relevant_trackers(id, &ctx, |tracker| {
-            tracker.record_metadata(id, tracker.path(id, &ctx), values);
+        self.for_relevant_trackers(id, &ctx, |tracker, scope| {
+            tracker.record_metadata(id, tracker.path(scope), values);
         })
     }
 
     fn on_enter(&self, id: &Id, ctx: Context<'_, S>) {
-        self.for_relevant_trackers(id, &ctx, |tracker| {
-            tracker.open(id, tracker.path(id, &ctx), SpanInfo::for_span(id, &ctx));
+        self.for_relevant_trackers(id, &ctx, |tracker, scope| {
+            tracker.open(id, tracker.path(scope), SpanInfo::for_span(id, &ctx));
         });
     }
 
     fn on_close(&self, id: Id, ctx: Context<'_, S>) {
-        self.for_relevant_trackers(&id, &ctx, |tracker| {
-            tracker.exit(&id, tracker.path(&id, &ctx), SystemTime::now())
+        self.for_relevant_trackers(&id, &ctx, |tracker, scope| {
+            tracker.exit(&id, tracker.path(scope), SystemTime::now())
         });
-        if let Some(tracker) = self.tracked_spans.read().unwrap().get(&id) {
+        if let Some(tracker) = self.tracked_spans.read().get(&id) {
             let _ = tracker.dump(self.settings());
         }
     }
@@ -779,7 +797,7 @@ where
 
 #[cfg(test)]
 mod test {
-    use crate::{width, Settings, SpanInfo, SpanTracker};
+    use crate::{width, Settings, SpanInfo, SpanTracker, TrackedMetadata};
     use std::iter;
     use std::ops::Add;
     use std::time::{Duration, UNIX_EPOCH};
@@ -822,6 +840,26 @@ mod test {
         String::from_utf8(out.clone()).unwrap()
     }
 
+    #[test]
+    fn render_metadata() {
+        let metadata = TrackedMetadata {
+            data: vec![("A".to_string(), "B".to_string()), ("c".to_string(), "d".to_string())]
+        };
+        let mut out = String::new();
+        metadata.write(&mut out, &Settings::default()).unwrap();
+        assert_eq!(out, "{A: B, c: d}");
+    }
+
+    #[test]
+    fn render_empty_metadata() {
+        let metadata = TrackedMetadata {
+            data: vec![]
+        };
+        let mut out = String::new();
+        metadata.write(&mut out, &Settings::default()).unwrap();
+        assert_eq!(out, "");
+    }
+
     #[test]
     fn render_correct_output() {
         let id_0 = Id::from_u64(1);
@@ -833,7 +871,7 @@ mod test {
             &id_0,
             iter::empty(),
             SpanInfo {
-                name: "test".to_string(),
+                name: "test",
                 start: interval_start,
                 end: None,
             },
@@ -843,7 +881,7 @@ mod test {
                 &id_1,
                 &mut [id_1.clone()].iter().cloned(),
                 SpanInfo {
-                    name: "nested".to_string(),
+                    name: "nested",
                     start: interval_start + Duration::from_secs(2),
                     end: None,
                 },
